# -*- coding:utf-8 -*-
# Author: Bingning Wang (王炳宁)
# Email: research@bingning.wang
# HomePage: www.bingning.wang
# Date: 2015/10/25 0025 上午 9:31
from public_functions import *
from joblib import delayed
data_dir = 'data/WikiQACorpus/'

word2id = {'UNK': 0}

wordCount = {}
sentence_pool = []
train_file_path = data_dir + 'WikiQA-train.txt'
test_file_path = data_dir + 'WikiQA-test.txt'
dev_file_path = data_dir + 'WikiQA-dev.txt'


def process_string(sent):
    return sent.lower()


def add_word_count(line):
    words = line.split(' ')
    for word in words:
        if word not in wordCount:
            wordCount[word] = 0
        wordCount[word] += 1


def get_word2id():
    dump_file(wordCount, data_dir + 'wordCount.obj')
    orderCount = sorted(wordCount.items(), lambda x, y: cmp(x[1], y[1]), reverse=True)
    lst = [str(x[0]) + '\t' + str(x[1]) for x in orderCount]
    write_list2file(lst, data_dir + 'wordCount.txt')
    for x in orderCount:
        word2id[x[0]] = len(word2id)
    dump_file(word2id, data_dir + 'word2id.obj')


def pre_process_data():
    one_data = {}
    with open(train_file_path) as f:
        for line in f:
            ss = line.strip().split('\t')
            question = process_string(ss[0])
            candidate_answer = process_string(ss[1])
            label = int(ss[2])
            if question not in one_data:
                one_data[question] = []
            one_data[question].append([candidate_answer, label])
    with open(dev_file_path) as f:
        for line in f:
            ss = line.strip().split('\t')
            question = process_string(ss[0])
            candidate_answer = process_string(ss[1])
            label = int(ss[2])
            if question not in one_data:
                one_data[question] = []
            one_data[question].append([candidate_answer, label])
    with open(test_file_path) as f:
        for line in f:
            ss = line.strip().split('\t')
            question = process_string(ss[0])
            candidate_answer = process_string(ss[1])
            label = int(ss[2])
            if question not in one_data:
                one_data[question] = []
            one_data[question].append([candidate_answer, label])
    # as we only consider the questions that has the right answer, we remove those that dont have the right answer
    for one in one_data:
        a_l = one_data[one]
        add_word_count(one)
        sentence_pool.extend([x[0] for x in a_l])  # we enhance the dataset
        [add_word_count(x[0]) for x in a_l]
    dump_file(sentence_pool, data_dir + 'sentence_pool.obj')
    print 'word count numbers:', len(wordCount)
    print 'sentence pool size:', len(sentence_pool)
    get_word2id()


def process_data():
    def map2id(sent):
        words = sent.strip().split(' ')
        out = []
        for word in words:
            if word in word2id:
                out.append(word2id[word])
            else:
                out.append(0)
        return out

    def generate_q_a_triplet(one_pool, add_random_neg=True):
        triplet_pool = []
        for one in one_pool:
            question = one[0]
            pos = []
            neg = []
            for q_a in one[1]:
                if q_a[1] == 1:
                    pos.append(q_a[0])
                else:
                    neg.append(q_a[0])
            if add_random_neg:
                neg.extend(np.random.choice(sentence_pool, 15).tolist())
            for one_pos in pos:
                for one_neg in neg:
                    triplet_pool.append([question, one_pos, one_neg])
        return triplet_pool

    def deal_one_data(filepath):
        one_data = {}
        with open(filepath) as f:
            for line in f:
                ss = line.strip().split('\t')
                question = process_string(ss[0])
                candidate_answer = process_string(ss[1])
                label = int(ss[2])
                if question not in one_data:
                    one_data[question] = []
                one_data[question].append([candidate_answer, label])
        # as we only consider the questions that has the right answer, we remove those that dont have the right answer
        output_data = []
        for one in one_data:
            a_l = one_data[one]
            if sum([x[1] for x in a_l]) > 0:
                output_data.append([one, a_l])
        return output_data

    train_output_data = deal_one_data(train_file_path)
    train_pool = generate_q_a_triplet(train_output_data)
    print 'train data size:', len(train_pool)
    train_data = [[map2id(x[0]), map2id(x[1]), map2id(x[2])] for x in train_pool]
    dump_file(train_data, data_dir + 'train.obj')
    dev_output_data = deal_one_data(dev_file_path)
    print 'dev data size:', len(dev_output_data)
    dev_data = []
    for one in dev_output_data:
        questions = map2id(one[0])
        candidates = [map2id(x[0]) for x in one[1]]
        labels = [x[1] for x in one[1]]
        dev_data.append([questions, zip(candidates, labels)])
    dump_file(dev_data, data_dir + 'dev.obj')

    test_output_data = deal_one_data(test_file_path)
    print 'test data size:', len(test_output_data)
    test_data = []
    for one in test_output_data:
        questions = map2id(one[0])
        candidates = [map2id(x[0]) for x in one[1]]
        labels = [x[1] for x in one[1]]
        test_data.append([questions, zip(candidates, labels)])
    dump_file(test_data, data_dir + 'test.obj')

if __name__ == '__main__':
    pre_process_data()
    process_data()
