# -*- coding:utf-8 -*-
# Author: Bingning Wang (王炳宁)
# Email: research@bingning.wang
# HomePage: www.bingning.wang
# Date: 2015/10/27 0025 上午 7:31
from RNN import MultiLayerRNN, GRU, MultiLayerGRUCONTEXT, MultiLayerGRUGATE
from modelBase import *


class IARNN(Model):
    def __init__(self,
                 **kwargs):
        Model.__init__(self, **kwargs)

    def build_model(self, is_test=False):
        question_input = T.imatrix('question')
        pos_input = T.imatrix('pos')
        neg_input = T.imatrix('neg')
        updated_word_embedding = theano.shared(self.word_embedding[:self.update_word_embedding_size, :],
                                               'word_embedding')
        fixed_word_embedding = theano.shared(self.word_embedding[self.update_word_embedding_size:, :],
                                             'fixed_word_embedding')
        word_embedding = T.concatenate([updated_word_embedding, fixed_word_embedding], axis=0)
        question_word_embedding = word_embedding[question_input]
        pos_word_embedding = word_embedding[pos_input]
        neg_word_embedding = word_embedding[neg_input]
        # first we derive the question representation
        neg_score, pos_score, parameter = self.get_score(question_word_embedding, pos_word_embedding,
                                                         neg_word_embedding)
        parameter.append(updated_word_embedding)

        loss = T.mean(T.nnet.relu(0.15 + neg_score - pos_score))
        updates = self.get_updates(loss, parameter)

        if is_test:
            function = get_function(inputs=[question_input, pos_input],
                                    outputs=pos_score)
        else:
            function = get_function(inputs=[question_input, pos_input, neg_input],
                                    outputs=loss,
                                    updates=updates
                                    )
        return [function, parameter]

    @NotImplementedError
    def get_score(self, question_word_embedding, pos_word_embedding, neg_word_embedding):
        pass


class IARNN_Word(IARNN):
    def __init__(self,
                 **kwargs):
        IARNN.__init__(self, **kwargs)

    def get_score(self, question_word_embedding, pos_word_embedding, neg_word_embedding):
        encoder = MultiLayerRNN(n_word_embedding=self.n_word_embedding, n_hidden=self.encoder_size, n_layer=1,
                                cell=GRU,
                                name='encoder')

        question_representations = encoder(question_word_embedding)

        w_question_attention = init_bias(2 * self.encoder_size, name='w_question_attention')
        w_answer_attention = init_bias(2 * self.encoder_size, name='w_answer_attention')
        question_word_score = T.nnet.softmax(T.dot(question_representations, w_question_attention))

        question_representation = T.batched_dot(question_word_score, question_representations)

        W_word_attention = init_parameter(2 * self.encoder_size, self.n_word_embedding, name='W_word_attention')

        q_w_representation = T.dot(question_representation, W_word_attention)

        def get_question_attentive_word_representation(word_representation):
            score = T.nnet.sigmoid(T.batched_dot(word_representation, q_w_representation))
            return score.dimshuffle(0, 1, 'x') * word_representation

        pos_word_attentive_representation = get_question_attentive_word_representation(pos_word_embedding)
        neg_word_attentive_representation = get_question_attentive_word_representation(neg_word_embedding)

        pos_representations = encoder(pos_word_attentive_representation)
        neg_representations = encoder(neg_word_attentive_representation)

        w_answer_attention = init_bias(2 * self.encoder_size, name='w_answer_attention')

        def get_rnn_output(representations):
            score = T.nnet.softmax(T.dot(representations, w_answer_attention))
            return T.batched_dot(score, representations)

        pos_representation = get_rnn_output(pos_representations)
        neg_representation = get_rnn_output(neg_representations)

        def get_score(representations):
            num = T.batched_dot(representations, question_representation)
            dom = representations.norm(L=2, axis=1) * question_representation.norm(L=2, axis=1)
            return num / dom

        pos_score = get_score(pos_representation)
        neg_score = get_score(neg_representation)
        parameter = encoder.parameter + [w_answer_attention, w_question_attention,
                                         W_word_attention]
        return neg_score, pos_score, parameter


class IARNN_Context(IARNN):
    def __init__(self,
                 **kwargs):
        IARNN.__init__(self, **kwargs)

    def get_score(self, question_word_embedding, pos_word_embedding, neg_word_embedding):
        encoder_question = MultiLayerRNN(n_word_embedding=self.n_word_embedding, n_hidden=self.encoder_size, n_layer=1,
                                         cell=GRU,
                                         name='encoder_question')
        question_representations = encoder_question(question_word_embedding)

        w_question_attention = init_bias(2 * self.encoder_size, name='w_question_attention')
        question_word_score = T.nnet.softmax(T.dot(question_representations, w_question_attention))

        question_representation = T.batched_dot(question_word_score, question_representations)

        encoder_answer = MultiLayerGRUCONTEXT(attention_representation=question_representation,
                                              n_attention=2 * self.encoder_size, n_word_embedding=self.n_word_embedding,
                                              n_hidden=self.encoder_size, n_layer=1,
                                              name='encoder_answer')

        pos_representations = encoder_answer(pos_word_embedding)
        neg_representations = encoder_answer(neg_word_embedding)

        w_answer_attention = init_bias(2 * self.encoder_size, name='w_answer_attention')

        def get_rnn_output(representations):
            score = T.nnet.softmax(T.dot(representations, w_answer_attention))
            return T.batched_dot(score, representations)

        pos_representation = get_rnn_output(pos_representations)
        neg_representation = get_rnn_output(neg_representations)

        def get_score(representations):
            num = T.batched_dot(representations, question_representation)
            dom = representations.norm(L=2, axis=1) * question_representation.norm(L=2, axis=1)
            return num / dom

        pos_score = get_score(pos_representation)
        neg_score = get_score(neg_representation)
        parameter = encoder_question.parameter + encoder_answer.parameter + [w_answer_attention, w_question_attention]
        return neg_score, pos_score, parameter


class IARNN_Gate(IARNN):
    def __init__(self,
                 **kwargs):
        IARNN.__init__(self, **kwargs)

    def get_score(self, question_word_embedding, pos_word_embedding, neg_word_embedding):
        encoder_question = MultiLayerRNN(n_word_embedding=self.n_word_embedding, n_hidden=self.encoder_size, n_layer=1,
                                         cell=GRU,
                                         name='encoder_question')
        question_representations = encoder_question(question_word_embedding)

        w_question_attention = init_bias(2 * self.encoder_size, name='w_question_attention')
        question_word_score = T.nnet.softmax(T.dot(question_representations, w_question_attention))

        question_representation = T.batched_dot(question_word_score, question_representations)

        encoder_answer = MultiLayerGRUGATE(attention_representation=question_representation,
                                           n_attention=2 * self.encoder_size, n_word_embedding=self.n_word_embedding,
                                           n_hidden=self.encoder_size, n_layer=1,
                                           name='encoder_answer')

        pos_representations = encoder_answer(pos_word_embedding)
        neg_representations = encoder_answer(neg_word_embedding)

        w_answer_attention = init_bias(2 * self.encoder_size, name='w_answer_attention')

        def get_rnn_output(representations):
            score = T.nnet.softmax(T.dot(representations, w_answer_attention))
            return T.batched_dot(score, representations)

        pos_representation = get_rnn_output(pos_representations)
        neg_representation = get_rnn_output(neg_representations)

        def get_score(representations):
            num = T.batched_dot(representations, question_representation)
            dom = representations.norm(L=2, axis=1) * question_representation.norm(L=2, axis=1)
            return num / dom

        pos_score = get_score(pos_representation)
        neg_score = get_score(neg_representation)
        parameter = encoder_question.parameter + encoder_answer.parameter + [w_answer_attention, w_question_attention]
        return neg_score, pos_score, parameter
