# -*- coding:utf-8 -*-
# Author: Bingning Wang (王炳宁)
# Email: research@bingning.wang
# HomePage: www.bingning.wang
# Date: 2017/10/14 0014 上午 10:59
from tqdm import tqdm

from ParameterUtils import save_parameter_with_name, load_parameters
from public_functions import *
from theanoUtils import *

glove_data_path = ''


class Model:
    def __init__(self,
                 encoder_size=128,
                 n_word_embedding=300,
                 update_word_embedding_size=100,
                 model_number=None,
                 optimizer='adamax',
                 batch_size=32,
                 l1=1.0e-7,
                 l2=1.0e-5,
                 ):
        self.encoder_size = encoder_size
        self.update_word_embedding_size = update_word_embedding_size
        self.n_word_embedding = n_word_embedding
        self.optimizer = optimizer
        self.batch_size = batch_size
        self.l2 = l2
        self.l1 = l1
        self.model_number = np.random.randint(1, 1e4) if model_number is None else model_number
        print 'model number', self.model_number
        self.modelDir = 'model/' + str(self.model_number)
        self.word2id = load_file('data/WikiQACorpus/word2id.obj')
        self.vocab_size = len(self.word2id)
        self.word_embedding = self.get_word_embedding(reload_=False)

    def get_word_embedding(self, reload_=False):
        if reload_:
            wordEmbedding = np.float32(np.random.normal(scale=0.4, size=(self.vocab_size, self.n_word_embedding)))
            number = 0
            print self.vocab_size
            with open(glove_data_path) as f:
                for line in tqdm(f):
                    s = line.split(' ')
                    word = s[0]
                    if word in self.word2id:
                        w_vecstr = np.array(s[1:self.n_word_embedding + 1], dtype='float32')
                        wordEmbedding[self.word2id[word]] = w_vecstr
                        number += 1
            print number
            dump_file(wordEmbedding, 'data/WikiQACorpus/word2vec.obj')
        else:
            wordEmbedding = load_file('data/WikiQACorpus/word2vec.obj')
        return wordEmbedding

    def get_updates(self, loss, parameter):
        loss = add_l1_l2_norm(self.l1, self.l2, loss, parameter)
        updates = get_update(self.optimizer, loss, parameter)
        return updates

    def dump_function(self, parameter, epoch):
        if not os.path.exists(self.modelDir):
            os.mkdir(self.modelDir)
        save_parameter_with_name(parameter, self.modelDir + '/parameter_' + str(epoch) + '.bin')

    @NotImplementedError
    def build_model(self):
        pass

    @NotImplementedError
    def train(self):
        pass

    def reload_parameter(self, parameter, epoch):
        load_parameters(parameter, self.modelDir + '/parameter_' + str(epoch) + '.bin')
